import processing.core.*; import processing.serial.*; import java.applet.*; import java.awt.*; import java.awt.image.*; import java.awt.event.*; import java.io.*; import java.net.*; import java.text.*; import java.util.*; import java.util.zip.*; public class roombacomm_tst extends PApplet {//
// RoombaComm User Interface
//
//  Tod E. Kurt, tod@todbot.com
//
//

RoombaComm roombaComm = null;

// if you set this to not -1, it will not do the port picker UI
//int roombaCommPort = -1;
int roombaCommPort = 3;
int sensor_update_every = 5;  // update every 5 frames
// maps dir values to roomba radius numbers (absolute value)
int[] dirmap = { 1500, 1000, 500, 250, 125, 60, 30, 15 };
int max_dir  = dirmap.length - 1;
int max_speed = 5;
float angle_fudge_factor = 0.25f;  // FIXME: why a divide-by-four?

int sensor_framecnt = 0;
int sdistance, sangle;
boolean safetyfault = false;
boolean sensorfault = false;

int c,d;
int octave = 3;

float rx,ry,rangle;

PFont fontA;
boolean toggleDraw = false;

// called by processing once on startup
public void setup()
{
  size(350,350);
  stroke(100);
  background(153);
  ellipseMode(CENTER_RADIUS);
  rectMode(CENTER);
  smooth();
  framerate(10);

  fontA = loadFont("CourierNewPSMT-24.vlw"); 
  textFont(fontA, 12);
  fill(0);
  text("use arrow keys to drive",  10,height-28);

  roombaComm = new RoombaComm(this);
  if( roombaCommPort >= 0 ) {
    if( roombaComm.connect(roombaCommPort) ) {
      println("connected to Roomba!");
      roombaComm.startup();
    }
  }
  
  rx = width/2;
  ry = height/2;
}

// FIXME: put this in RoombaComm?
//
// Present the user with a serial port selector
//
public void roombaCommSetup()
{
  int listlen = roombaComm.portlist.length;
  textFont(fontA, 12);
  background(153); fill(0); stroke(100);
  text("click a port:", 10,10);
  for( int i=0; i< listlen; i++ ) {
    text(i+": "+roombaComm.portlist[i], 10,20+10*i);
  }
  int s = (mouseY-15) / 10;
  if( s >= listlen || s < 0 ) return;
  rect(2,15+10*s, 10,8);
  if( mousePressed ) {
    line(10,20+10*s, width-10, 20+10*s);
    if( roombaComm.connect( s ) ) {
      roombaComm.startup();
      roombaCommPort = s;
    }
  }
}


public void helpScreen()
{
  
}

public void drawGridlines()
{
  int gdx = 50;
  stroke(100);    strokeWeight(1); 
  for( int i=0; i< 10; i++ ) {
    line(0, i*gdx, width, i*gdx);
    line(i*gdx,0, i*gdx, height);
  }
}

public void drawRoomba()
{
  float radius = 20.0f;

  if( toggleDraw ) {
//    image( roombapic, -radius,-radius);
  }
  else {
    noStroke();
    fill(45);
    ellipse(0, 0, radius, radius);
    fill(70);
    ellipse(0, 0, radius-3, radius-3);
    fill(160);
    rect(-radius+12,0, 15,4);
    fill(100);
    rect(radius/2,0, 4,4);
  }
}

// called by Processing 'framerate' times a second
public void draw()
{
  if( roombaComm.connected == false ) { roombaCommSetup(); return; }

  textFont(fontA, 12);
  fill(0);  background(153); stroke(100);

  if( ++sensor_framecnt == sensor_update_every ) {
    sensorfault = ! roombaComm.updateSensors();
    sensor_framecnt=0;
    sdistance = roombaComm.distance();
    sangle    = roombaComm.angle();
    //sangle_dxeg= int((360 * sangle)/(258*PI));
    println("update at "+millis());
  }

  drawGridlines();

  // draw status panel
  text("use arrow keys to drive",  8,height-28);
  text("space to stop, / to reset",8,height-18);
  text("asdf.. music keyboard",    8,height-8);
  text("  status: "+ (sensorfault?"no roomba":(safetyfault?"safety fault":"OK")),8, 8);
  text("distance: "+sdistance, width-130, height-18);
  text("   angle: "+sangle,    width-130, height-8);

  // draw joystick
  strokeWeight(4);   // Thicker
  int jposx = width/2, jposy = height/2; 
  line( jposx, jposy, jposx - (d*5), jposy - (c*5) );
  
  float da = ((2.0f * sangle)/258.0f) * angle_fudge_factor;
  float dx = sdistance * cos(da) / 10;
  float dy = sdistance * sin(da) / 10;
  rx = rx + dx;
  ry = ry + dy;
  rangle = rangle - da;
  rangle %= TWO_PI;
  println("sdist:"+sdistance+" sangle:"+sangle+" dx:"+dx+" dy:"+dy+" rangle:"+rangle);
  
  // torroidial mapping, sir                                                    
  if( rx > width ) rx = 0; if( rx < 0 ) rx = width;
  if( ry > width ) ry = 0; if( ry < 0 ) ry = height;

  if( roombaComm.safetyFault ) {
    textFont(fontA, 36);
    text("SAFETY FAULT!", 20,height/3);
  } 
  else {
    roombaComm.reset();
  }

  translate(rx,ry);
  rotate(rangle);
  drawRoomba();


    
}


public void keyPressed()
{
  int note=-1;
  boolean movekey = false;
  if( key == CODED ) {
    if( keyCode == UP ) {
      c++; movekey=true;
    } else if( keyCode == DOWN ) {
      c--; movekey=true;
    } else if( keyCode == LEFT ) {
      d++; movekey=true;
    } else if( keyCode == RIGHT ) {
      d--; movekey=true;
    }
  }
  else if( key == ' ' ) {
    roombaComm.stop();
    c = 0; d = 0; 
  }
  else if( key == '/' ) {
    roombaComm.reset();  // in case we pick the roomba up and it shifts to passive mode
    c = 0; d = 0;
  }
  else {
    switch( key ) {
      // pseudo-keyboard to play notes on
      case 'a': note = 0;  break; 
      case 'w': note = 1;  break; 
      case 's': note = 2;  break; 
      case 'e': note = 3;  break; 
      case 'd': note = 4;  break; 
      case 'f': note = 5;  break; 
      case 't': note = 6;  break; 
      case 'g': note = 7;  break; 
      case 'y': note = 8;  break; 
      case 'h': note = 9;  break; 
      case 'u': note =10;  break; 
      case 'j': note =11;  break; 
      case 'k': note =12;  break; 
      case 'o': note =13;  break; 
      case 'l': note =14;  break; 
      case 'z': octave--;  break;  // change octaves
      case 'x': octave++;  break;  // change octaves
    }
  }
  
  // we actually hit a note key
  if( note >= 0 ) {
    octave = (octave < 1) ? 1 : (octave > 7) ? 7 : octave;
    roombaComm.playNote( note+(octave*12) + 36, 10);
  }

  if( movekey ) {
    if( c>max_speed ) c=max_speed; if( c<-max_speed ) c=-max_speed;
    if( d>max_dir )   d=max_dir;   if( d<-max_dir )   d=-max_dir;
    
    println("c:"+c+", d:"+d);
    
    if( c == 0 && d == 0 ) {
      roombaComm.stop();
    } else if( d==0 ) {
      roombaComm.goStraight( c*75 );
    } else if( c==0 ) {
      if( d > 0 )
        roombaComm.spinLeft( d*50 );
      else
        roombaComm.spinRight( -d*50 );
    } else {
      if( d>0 )
        roombaComm.drive( c*75, dirmap[d] );
      else 
        roombaComm.drive( c*75, -dirmap[-d] );
    }
  }
  
  
}

//
// blah blah
//



// -------------------------------------------------------------------------
//
// -------------------------------------------------------------------------
class RoombaComm
{
  PApplet parent;
  Serial port;
  String[] portlist;
  boolean connected = false;
  boolean sensorsValid = false;
  boolean safetyFault = false;
  byte[] sensor_bytes = new byte[26];

  // constructor
  RoombaComm(PApplet p) {
    portlist = Serial.list();
    println(portlist);    // List all the available serial ports:
    parent = p;
  }
  
  //
  //
  //
  public boolean connect(int portnum) {
    String portname = portlist[portnum];
    println("connecting to port '"+portname+"'");
    port = new Serial(parent, portname, 57600);
    if( port != null && port.port != null ) // about the only way I know to do this
      connected = true;
    else  
      connected = false;
    sensorsValid = false;
    return connected;
  }
  
  //
  // toggles DD line via serial port DTR
  //
  public void wakeup() {
    // unimplementable until DTR control is available  
    // can probably do:
    // port.port.setDTR(false);
    //  try { Thread.sleep(500); } 
    // port.port.setDTR(true);
  }
  
  //
  //
  //
 public void startup() {
    port.write( START );
    port.write( CONTROL );
  }
  
  // FIXME: put this in RoombaComm?
  public void reset()
  {
    startup();
    updateSensors();
  }

  //
  // move the roomba via the low-level velocity + radius method
  //
  public void drive( int velocity, int radius ) {
    byte cmd[] = { DRIVE, PApplet.toByte(velocity >>> 8), PApplet.toByte(velocity & 0xff), 
                           PApplet.toByte(radius  >>> 8), PApplet.toByte(radius   & 0xff) };
    println("drive: "+hex(cmd[0])+","+hex(cmd[1])+","+hex(cmd[2])+","+hex(cmd[3])+","+hex(cmd[4]));
    port.write(cmd);
  }
  
  //
  // higher-level movement commands 
  // 
  public void stop() {
    drive( 0, 0 );
  }
  public void spinLeft( int speed ) {
    drive( speed, 1 ); 
  }
  public void spinRight( int speed ) {
    drive( speed, -1 ); 
  }
  public void spinLeft() {
    spinLeft( 100 );  
  }
  public void spinRight() {
    spinRight( 100 );  
  }
  public void goStraight( int speed ) {
    drive( speed, 0x8000 );
  }
  public void goStraight() {
     goStraight( 200 );
  }
  
  // turns on/off the various LEDs
  public void setLEDs( boolean status_green, boolean status_red, 
                boolean spot, boolean clean, boolean max, boolean dirt, 
                int power_color, int power_intensity ) {
    int v = (status_green?0x20:0) | (status_red?0x10:0) | (spot?0x08:0) | (clean?0x04:0) | (max?0x02:0) | (dirt?0x01:0);
    println("leds: "+binary(v));
    byte cmd[] = { LEDS, PApplet.toByte(v), PApplet.toByte(power_color), PApplet.toByte(power_intensity) };
    port.write(cmd);
  }
  
  // turns on/off the non-drive motors
  public void setMotors( boolean mainbrush, boolean vacuum, boolean sidebrush ) {
    port.write( MOTORS );
    port.write( PApplet.toByte( (mainbrush?0x04:0) | (vacuum?0x02:0) | (sidebrush?0x01:0) ) );
  }
  
  // plays a note via the hacky method of defining a one-note song & playing it
  // uses up song slot 15
  public void playNote( int note, int duration ) {
    println("playnote: "+note+":"+duration);
    byte cmd[] = { SONG, 15, 1, PApplet.toByte(note), PApplet.toByte(duration),
                    PLAY, 15 };
    port.write(cmd);
  }
  
  //
  public boolean updateSensors() {
    byte cmd[] = { SENSORS, 0 };
    port.write(cmd);
    sensorsValid = false;  // assume the worst
    int n = port.available();
    if( n >= 26) {                       // there are enough bytes to read
      n = port.readBytes(sensor_bytes);
      if( n==26 ) {                      // did we get enough?
        sensorsValid = true;             // then everything's good, otherwise bad
        computeSafetyFault();
      }
    } else {
      println("only "+n+" bytes available, not updating sensors");
    }
    return sensorsValid;
  }
  
  // if true, indicates we had an event that took the Roomba out of safe mode
  // called on every successful updateSensors
  // use the public boolean 'safety_fault' in your code
  public boolean computeSafetyFault() {
    safetyFault = (bumps_wheeldrops() & WHEELDROP_MASK) != 0 ||
            cliff_left() == 1  || cliff_frontleft() == 1 ||
            cliff_right() == 1 || cliff_frontright() == 1 ;
    return safetyFault;
  }

    
  public int bumps_wheeldrops() {
    return sensor_bytes[0];
  }
  public int wall() {
    return sensor_bytes[1];
  }
  public int cliff_left() {
    return sensor_bytes[2];
  }
  public int cliff_frontleft() {
    return sensor_bytes[3];
  }
  public int cliff_frontright() {
    return sensor_bytes[4];
  }
  public int cliff_right() {
    return sensor_bytes[5];
  }
  public int virtual_wall() {
    return sensor_bytes[6];
  }
  public int motor_overcurrents() {
    return sensor_bytes[7];
  }
  public int dirt_left() {
    return sensor_bytes[8];
  }
  public int dirt_right() {
    return sensor_bytes[9];
  }
  
  public int remote_opcode() {
    return sensor_bytes[10];
  }
  public int buttons() {
    return sensor_bytes[11];
  }
  public int distance() {
    //println("distH:"+sensor_bytes[12] + ", distL:"+sensor_bytes[13]);
    return sensor_bytes[12] << 8 | sensor_bytes[13];
  }
  public int angle() {
    return sensor_bytes[14] << 8 | sensor_bytes[15];
  }
  
  public int charging_state() {
    return sensor_bytes[16];
  }
  public int voltage() {
    return sensor_bytes[17] << 8 | sensor_bytes[18];    
  }
  public int current() {
    return sensor_bytes[19] << 8 | sensor_bytes[20];
  }
  public int temperature() {
    return sensor_bytes[21]; 
  }
  public int charge() {
    return sensor_bytes[22] << 8 | sensor_bytes[23];
  }
  public int capacity() {
    return sensor_bytes[24] << 8 | sensor_bytes[25];
  }

    
  // Roomba SCI opcodes
  // the dumb byte...byte() crap is because processing/java bytes are signed
  byte START   =  PApplet.toByte(128);  // 0
  byte BAUD    =  PApplet.toByte(129);  // 1
  byte CONTROL =  PApplet.toByte(130);  // 0
  byte SAFE    =  PApplet.toByte(131);  // 0
  byte FULL    =  PApplet.toByte(132);  // 0
  byte POWER   =  PApplet.toByte(133);  // 0
  byte SPOT    =  PApplet.toByte(134);  // 0
  byte CLEAN   =  PApplet.toByte(135);  // 0
  byte MAX     =  PApplet.toByte(136);  // 0
  byte DRIVE   =  PApplet.toByte(137);  // 4
  byte MOTORS  =  PApplet.toByte(138);  // 1
  byte LEDS    =  PApplet.toByte(139);  // 3
  byte SONG    =  PApplet.toByte(140);  // 2N+2
  byte PLAY    =  PApplet.toByte(141);  // 1
  byte SENSORS =  PApplet.toByte(142);  // 1
  byte DOCK    =  PApplet.toByte(143);  // 0

  // bitmasks for various thingems
  byte WHEELDROP_MASK = 0x1C;
  
}
static public void main(String args[]) {   PApplet.main(new String[] { "roombacomm_tst" });}}