//
// blah blah
//

import processing.serial.*;

// -------------------------------------------------------------------------
//
// -------------------------------------------------------------------------
class RoombaComm
{
  PApplet parent;
  Serial port;
  String[] portlist;
  boolean connected = false;
  boolean sensorsValid = false;
  boolean safetyFault = false;
  byte[] sensor_bytes = new byte[26];

  // constructor
  RoombaComm(PApplet p) {
    portlist = Serial.list();
    println(portlist);    // List all the available serial ports:
    parent = p;
  }
  
  //
  //
  //
  boolean connect(int portnum) {
    String portname = portlist[portnum];
    println("connecting to port '"+portname+"'");
    port = new Serial(parent, portname, 57600);
    if( port != null && port.port != null ) // about the only way I know to do this
      connected = true;
    else  
      connected = false;
    sensorsValid = false;
    return connected;
  }
  
  //
  // toggles DD line via serial port DTR
  //
  void wakeup() {
    // unimplementable until DTR control is available  
    // can probably do:
    // port.port.setDTR(false);
    //  try { Thread.sleep(500); } 
    // port.port.setDTR(true);
  }
  
  //
  //
  //
 void startup() {
    port.write( START );
    port.write( CONTROL );
  }
  
  // FIXME: put this in RoombaComm?
  void reset()
  {
    startup();
    updateSensors();
  }

  //
  // move the roomba via the low-level velocity + radius method
  //
  void drive( int velocity, int radius ) {
    byte cmd[] = { DRIVE, byte(velocity >>> 8), byte(velocity & 0xff), 
                           byte(radius  >>> 8), byte(radius   & 0xff) };
    println("drive: "+hex(cmd[0])+","+hex(cmd[1])+","+hex(cmd[2])+","+hex(cmd[3])+","+hex(cmd[4]));
    port.write(cmd);
  }
  
  //
  // higher-level movement commands 
  // 
  void stop() {
    drive( 0, 0 );
  }
  void spinLeft( int speed ) {
    drive( speed, 1 ); 
  }
  void spinRight( int speed ) {
    drive( speed, -1 ); 
  }
  void spinLeft() {
    spinLeft( 100 );  
  }
  void spinRight() {
    spinRight( 100 );  
  }
  void goStraight( int speed ) {
    drive( speed, 0x8000 );
  }
  void goStraight() {
     goStraight( 200 );
  }
  
  // turns on/off the various LEDs
  void setLEDs( boolean status_green, boolean status_red, 
                boolean spot, boolean clean, boolean max, boolean dirt, 
                int power_color, int power_intensity ) {
    int v = (status_green?0x20:0) | (status_red?0x10:0) | (spot?0x08:0) | (clean?0x04:0) | (max?0x02:0) | (dirt?0x01:0);
    println("leds: "+binary(v));
    byte cmd[] = { LEDS, byte(v), byte(power_color), byte(power_intensity) };
    port.write(cmd);
  }
  
  // turns on/off the non-drive motors
  void setMotors( boolean mainbrush, boolean vacuum, boolean sidebrush ) {
    port.write( MOTORS );
    port.write( byte( (mainbrush?0x04:0) | (vacuum?0x02:0) | (sidebrush?0x01:0) ) );
  }
  
  // plays a note via the hacky method of defining a one-note song & playing it
  // uses up song slot 15
  void playNote( int note, int duration ) {
    println("playnote: "+note+":"+duration);
    byte cmd[] = { SONG, 15, 1, byte(note), byte(duration),
                    PLAY, 15 };
    port.write(cmd);
  }
  
  //
  boolean updateSensors() {
    byte cmd[] = { SENSORS, 0 };
    port.write(cmd);
    sensorsValid = false;  // assume the worst
    int n = port.available();
    if( n >= 26) {                       // there are enough bytes to read
      n = port.readBytes(sensor_bytes);
      if( n==26 ) {                      // did we get enough?
        sensorsValid = true;             // then everything's good, otherwise bad
        computeSafetyFault();
      }
    } else {
      println("only "+n+" bytes available, not updating sensors");
    }
    return sensorsValid;
  }
  
  // if true, indicates we had an event that took the Roomba out of safe mode
  // called on every successful updateSensors
  // use the public boolean 'safety_fault' in your code
  boolean computeSafetyFault() {
    safetyFault = (bumps_wheeldrops() & WHEELDROP_MASK) != 0 ||
            cliff_left() == 1  || cliff_frontleft() == 1 ||
            cliff_right() == 1 || cliff_frontright() == 1 ;
    return safetyFault;
  }

    
  int bumps_wheeldrops() {
    return sensor_bytes[0];
  }
  int wall() {
    return sensor_bytes[1];
  }
  int cliff_left() {
    return sensor_bytes[2];
  }
  int cliff_frontleft() {
    return sensor_bytes[3];
  }
  int cliff_frontright() {
    return sensor_bytes[4];
  }
  int cliff_right() {
    return sensor_bytes[5];
  }
  int virtual_wall() {
    return sensor_bytes[6];
  }
  int motor_overcurrents() {
    return sensor_bytes[7];
  }
  int dirt_left() {
    return sensor_bytes[8];
  }
  int dirt_right() {
    return sensor_bytes[9];
  }
  
  int remote_opcode() {
    return sensor_bytes[10];
  }
  int buttons() {
    return sensor_bytes[11];
  }
  int distance() {
    //println("distH:"+sensor_bytes[12] + ", distL:"+sensor_bytes[13]);
    return sensor_bytes[12] << 8 | sensor_bytes[13];
  }
  int angle() {
    return sensor_bytes[14] << 8 | sensor_bytes[15];
  }
  
  int charging_state() {
    return sensor_bytes[16];
  }
  int voltage() {
    return sensor_bytes[17] << 8 | sensor_bytes[18];    
  }
  int current() {
    return sensor_bytes[19] << 8 | sensor_bytes[20];
  }
  int temperature() {
    return sensor_bytes[21]; 
  }
  int charge() {
    return sensor_bytes[22] << 8 | sensor_bytes[23];
  }
  int capacity() {
    return sensor_bytes[24] << 8 | sensor_bytes[25];
  }

    
  // Roomba SCI opcodes
  // the dumb byte...byte() crap is because processing/java bytes are signed
  byte START   =  byte(128);  // 0
  byte BAUD    =  byte(129);  // 1
  byte CONTROL =  byte(130);  // 0
  byte SAFE    =  byte(131);  // 0
  byte FULL    =  byte(132);  // 0
  byte POWER   =  byte(133);  // 0
  byte SPOT    =  byte(134);  // 0
  byte CLEAN   =  byte(135);  // 0
  byte MAX     =  byte(136);  // 0
  byte DRIVE   =  byte(137);  // 4
  byte MOTORS  =  byte(138);  // 1
  byte LEDS    =  byte(139);  // 3
  byte SONG    =  byte(140);  // 2N+2
  byte PLAY    =  byte(141);  // 1
  byte SENSORS =  byte(142);  // 1
  byte DOCK    =  byte(143);  // 0

  // bitmasks for various thingems
  byte WHEELDROP_MASK = 0x1C;
  
}
